package ru.dbotthepony.mc.otm.data

import com.google.gson.JsonObject
import com.google.gson.JsonSyntaxException
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.JsonOps
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeSerializer
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.util.readBinaryJsonWithCodecIndirect
import ru.dbotthepony.mc.otm.core.util.writeBinaryJsonWithCodec
import java.util.*
import kotlin.NoSuchElementException
import kotlin.concurrent.getOrSet

class Codec2RecipeSerializer<S : Recipe<*>>(val empty: S?, val codec: Codec<S>) : Codec<S> by codec, RecipeSerializer<S> {
	constructor(supplier: (() -> ResourceLocation) -> Codec<S>) : this(null, supplier.invoke(::context))
	constructor(empty: S, supplier: (() -> ResourceLocation) -> Codec<S>) : this(empty, supplier.invoke(::context))

	override fun fromJson(p_44103_: ResourceLocation, p_44104_: JsonObject): S {
		try {
			deck.getOrSet(::LinkedList).addLast(p_44103_)

			return codec.decode(JsonOps.INSTANCE, p_44104_).get().map(
				{
					it.first
				},
				{
					empty ?: throw JsonSyntaxException("Failed to deserialize recipe from JSON: ${it.message()}")
				}
			)
		} finally {
			deck.get().removeLast()
		}
	}

	override fun fromNetwork(p_44105_: ResourceLocation, p_44106_: FriendlyByteBuf): S? {
		try {
			deck.getOrSet(::LinkedList).addLast(p_44105_)

			return p_44106_.readBinaryJsonWithCodecIndirect(codec)
				.resultOrPartial { LOGGER.error("Failed to read recipe $p_44105_ from network: $it") }.orElse(null)
		} finally {
			deck.get().removeLast()
		}
	}

	override fun toNetwork(p_44101_: FriendlyByteBuf, p_44102_: S) {
		p_44101_.writeBinaryJsonWithCodec(codec, p_44102_)
	}

	companion object : Codec<ResourceLocation> {
		private val deck = ThreadLocal<LinkedList<ResourceLocation>>()

		private fun context(): ResourceLocation {
			val deck = deck.getOrSet(::LinkedList)

			if (deck.isEmpty()) {
				throw NoSuchElementException("Context stack is empty")
			} else {
				return deck.last
			}
		}

		override fun <T : Any> encode(input: ResourceLocation, ops: DynamicOps<T>, prefix: T): DataResult<T> {
			return DataResult.success(ops.empty())
		}

		override fun <T : Any> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<ResourceLocation, T>> {
			val deck = deck.getOrSet(::LinkedList)

			if (deck.isEmpty()) {
				return DataResult.error { "Attempt to use recipe serializer codec ResourceLocation' hack outside Codec2RecipeSerializer" }
			} else {
				return DataResult.success(Pair(deck.last, ops.empty()))
			}
		}

		private val LOGGER = LogManager.getLogger()
	}
}
